from __future__ import print_function
import sys
import time
import os.path as osp
import os
from PIL import Image
import numpy as np
from collections import defaultdict

from .Dataset import Dataset

from ..utils.utils import measure_time
from ..utils.re_ranking import re_ranking
from ..utils.metric import cmc, mean_ap
from ..utils.dataset_utils import parse_im_name
from ..utils.distance import normalize
from ..utils.distance import compute_dist

import math


class TestSet(Dataset):
  """
  Args:
    extract_feat_func: a function to extract features. It takes a batch of
      images and returns a batch of features.
    marks: a list, each element e denoting whether the image is from 
      query (e == 0), or
      gallery (e == 1), or 
      multi query (e == 2) set
  """

  def __init__(
      self,
      im_dir=None,
      im_names=None,
      marks=None,
      extract_feat_func=None,
      single_gallery_shot=None,
      first_match_break=None,
      **kwargs):

    super(TestSet, self).__init__(dataset_size=len(im_names), **kwargs)

    # The im dir of all images
    self.im_dir = im_dir
    self.im_names = im_names
    self.marks = marks
    self.extract_feat_func = extract_feat_func
    self.single_gallery_shot = single_gallery_shot
    self.first_match_break = first_match_break

    window_f = os.path.abspath('/home/bing/Dataset/VehicleID/window-VehicleID.txt')
    lines = {}
    with open(window_f, 'r') as f:
      lines = {l.strip().split(' ')[0]: l.strip().split(' ')[1:6] for l in f if
               not lines.has_key(l.strip().split(' ')[0])}
    self.window = lines

    light_f = os.path.abspath('/home/bing/Dataset/VehicleID/light_pro.txt')
    lines = {}
    with open(light_f, 'r') as f:
      # lines = [[l.strip().split(' ')[0]]+l.strip().split(' ')[1:6] for l in f]
      for l in f:
        items = l.strip().split(' ')
        if float(items[1]) < 0.5:
          continue
        if lines.has_key(items[0]):
          lines[items[0]].append(map(float, items[2:]))
        else:
          lines[items[0]] = []
          lines[items[0]].append(map(float, items[2:]))
    self.light = lines

  def set_feat_func(self, extract_feat_func):
    self.extract_feat_func = extract_feat_func

  def get_sample(self, ptr):
    im_name = self.im_names[ptr]
    im_path = osp.join(self.im_dir, im_name)
    im = np.asarray(Image.open(im_path))

    # window
    if self.window.has_key(im_name[:-4]):
      box1 =  map(float, self.window[im_name[:-4]][1:])
    else:
      box1 = -1
    # lights
    if self.light.has_key(im_name[:-4]):
      box2 = self.light[im_name[:-4]]
    else:
      box2 = -1

    box1, box2, box3, box4 = self.box_generate(im, box1, box2)

    im, _ = self.pre_process_im(im)
    id = parse_im_name(self.im_names[ptr], 'id')
    tag = parse_im_name(self.im_names[ptr], 'tag')
    # denoting whether the im is from query, gallery, or multi query set
    mark = self.marks[ptr]
    return im, id, tag, im_name, box1, box2, box3, box4,mark

  def box_generate(self,im,box1_,box2_):
    """
    :param box1s: [N,4] or [N, None]
    :param box2s: [N,k,5] or [N,None]
    :return:
    box1:[N,4]for window
    box2:[N,4]for left light
    box3:[N,4]for right light
    box4:[N,4]for brand
    """
    if box1_ ==-1:
      box1 = np.array([0,3,15,7])
    else:
      # x [0,256],x/16 [0,16]
      x1, y1, x2, y2 = box1_
      x1 = float(x1) / im.shape[1] * 256
      x2 = float(x2) / im.shape[1] * 256
      y1 = float(y1) / im.shape[0] * 256
      y2 = float(y2) / im.shape[0] * 256

      xtl = int(x1 / 16)
      xbr = int(math.ceil(x2 / 16))
      ytl = int(y1 / 16)
      ybr = int(math.ceil(y2 / 16))
      if not (xtl<xbr and ytl < ybr):#some error in window.txt,im 835_-1_-1_0009795_
        box1 = np.array([0, 3, 15, 7])
      else:
        box1 = np.array([xtl, ytl, xbr, ybr])
    if box2_ == -1:
      box2 = np.array([0, 11, 5, 16])
      box3 = np.array([11, 11, 16, 16])
      box4 = np.array([5, 11, 11, 16])
    elif len(box2_) == 1:
      x1, y1, x2, y2 = box2_[0]
      x1 = float(x1) / im.shape[1] * 256
      x2 = float(x2) / im.shape[1] * 256
      y1 = float(y1) / im.shape[0] * 256
      y2 = float(y2) / im.shape[0] * 256

      ytl = int(x1 / 16)
      ybr = int(math.ceil(x2 / 16))
      xtl = int(y1 / 16)
      xbr = int(math.ceil(y2 / 16))
      box2 = np.array([xtl, ytl, xbr, ybr])
      box3 = np.array([xtl, ytl, xbr, ybr])
      box4 = np.array([5, 11, 11, 16])
    else:
      x1, y1, x2, y2 = box2_[0]
      x1 = float(x1) / im.shape[1] * 256
      x2 = float(x2) / im.shape[1] * 256
      y1 = float(y1) / im.shape[0] * 256
      y2 = float(y2) / im.shape[0] * 256

      xtl1 = int(x1 / 16)
      xbr1 = int(math.ceil(x2 / 16))
      ytl1 = int(y1 / 16)
      ybr1 = int(math.ceil(y2 / 16))
      assert xbr1 <= 16 and ybr1 <= 16, 'xbr1<16'
      box2 = np.array([xtl1, ytl1, xbr1, ybr1])
      x1, y1, x2, y2 = box2_[1]
      x1 = float(x1) / im.shape[1] * 256
      x2 = float(x2) / im.shape[1] * 256
      y1 = float(y1) / im.shape[0] * 256
      y2 = float(y2) / im.shape[0] * 256
      xtl0 = int(x1 / 16)
      xbr0 = int(math.ceil(x2 / 16))
      ytl0 = int(y1 / 16)
      ybr0 = int(math.ceil(y2 / 16))
      assert xbr0 <= 16 and ybr0 <= 16, 'xbr0<16'
      box3 = np.array([xtl0, ytl0, xbr0, ybr0])

      xtl = xbr0 if xbr0 < xbr1 else xbr1
      xbr = xtl0 if xtl0 > xtl1 else xtl1
      ytl = ytl1 if ytl1 < ytl0 else ytl0
      ybr = ybr1 if ybr1 > ybr0 else ybr0

      if not (xtl < xbr and ytl < ybr):
        box4 = np.array([5, 11, 11, 16])
      else:
        box4 = np.array([xtl, ytl, xbr, ybr])
    return box1,box2,box3,box4

  def next_batch(self):
    if self.epoch_done and self.shuffle:
      self.prng.shuffle(self.im_names)
    samples, self.epoch_done = self.prefetcher.next_batch()
    im_list, ids, tags, im_names, box1, box2, box3, box4,marks = zip(*samples)
    # Transform the list into a numpy array with shape [N, ...]
    ims = np.stack(im_list, axis=0)
    ids = np.array(ids)
    tags = np.array(tags)
    im_names = np.array(im_names)
    marks = np.array(marks)
    box1 = np.vstack(box1)
    box2 = np.vstack(box2)
    box3 = np.vstack(box3)
    box4 = np.vstack(box4)

    return ims, ids, tags, im_names, box1, box2, box3, box4,marks, self.epoch_done

  def extract_feat(self, normalize_feat, verbose=True):
    """Extract the features of the whole image set.
    Args:
      normalize_feat: True or False, whether to normalize feature to unit length
      verbose: whether to print the progress of extracting feature
    Returns:
      feat: numpy array with shape [N, C]
      ids: numpy array with shape [N]
      tags: numpy array with shape [N]
      im_names: numpy array with shape [N]
      marks: numpy array with shape [N]
    """
    feat, ids, tags, im_names, box1, box2, box3, box4,marks = [], [], [], [], [],[],[],[],[]
    done = False
    step = 0
    printed = False
    st = time.time()
    last_time = time.time()
    while not done:
      ims_, ids_, tags_, im_names_, box1_, box2_, box3_, box4_,marks_, done = self.next_batch()
      feat_ = self.extract_feat_func(ims_,box1_, box2_, box3_, box4_)
      feat.append(feat_)
      ids.append(ids_)
      tags.append(tags_)
      im_names.append(im_names_)
      marks.append(marks_)
      box1.append(box1_)
      box2.append(box2_)
      box3.append(box3_)
      box4.append(box4_)

      if verbose:
        # Print the progress of extracting feature
        total_batches = (self.prefetcher.dataset_size
                         // self.prefetcher.batch_size + 1)
        step += 1
        if step % 20 == 0:
          if not printed:
            printed = True
          else:
            # Clean the current line
            sys.stdout.write("\033[F\033[K")
          print('{}/{} batches done, +{:.2f}s, total {:.2f}s'
                .format(step, total_batches,
                        time.time() - last_time, time.time() - st))
          last_time = time.time()

    feat = np.vstack(feat)
    ids = np.hstack(ids)
    tags = np.hstack(tags)
    im_names = np.hstack(im_names)
    marks = np.hstack(marks)
    box1 = np.vstack(box1)
    box2 = np.vstack(box2)
    box3 = np.vstack(box3)
    box4 = np.vstack(box4)
    if normalize_feat:
      feat = normalize(feat, axis=1)
    return feat, ids, tags, im_names,marks

  def eval(
      self,
      normalize_feat=True,
      to_re_rank=False,
      pool_type='average',
      verbose=True):

    """Evaluate using metric CMC and mAP.
    Args:
      normalize_feat: whether to normalize features before computing distance
      to_re_rank: whether to also report re-ranking scores
      pool_type: 'average' or 'max', only for multi-query case
      verbose: whether to print the intermediate information
    """

    with measure_time('Extracting feature...', verbose=verbose):
      feat, ids, tags, im_names, marks = self.extract_feat(
        normalize_feat, verbose)

    # query, gallery, multi-query indices
    #q_inds = marks == 2
    #g_inds = marks == 3
    s_q_inds = marks == 0
    s_g_inds = marks == 1
    m_q_inds = marks == 2
    m_g_inds = marks == 3
    l_q_inds = marks == 4
    l_g_inds = marks == 5

    # A helper function just for avoiding code duplication.
    def compute_score(
        dist_mat,
        query_ids=ids[s_q_inds],
        gallery_ids=ids[s_g_inds],
        query_tags=tags[s_q_inds],
        gallery_tags=tags[s_g_inds]):
      # Compute mean AP
      mAP = mean_ap(
        distmat=dist_mat,
        query_ids=query_ids, gallery_ids=gallery_ids,
        query_tags=query_tags, gallery_tags=gallery_tags)
      # Compute CMC scores
      cmc_scores = cmc(
        distmat=dist_mat,
        query_ids=query_ids, gallery_ids=gallery_ids,
        query_tags=query_tags, gallery_tags=gallery_tags,
        single_gallery_shot=self.single_gallery_shot,
        first_match_break=self.first_match_break,
        topk=10)
      return mAP, cmc_scores

    def print_scores(mAP, cmc_scores):
      print('[mAP: {:5.2%}], [cmc1: {:5.2%}], [cmc5: {:5.2%}], [cmc10: {:5.2%}]'
            .format(mAP, *cmc_scores[[0, 4, 9]]))


    ################
    # Single Query #
    ################

    with measure_time('Computing distance...', verbose=verbose):
      # query-gallery distance
      q_g_dist = compute_dist(feat[s_q_inds], feat[s_g_inds], type='euclidean')
      s_indices = np.argsort(q_g_dist, axis=1)
      s_q_names = im_names[s_q_inds]
      s_g_names = im_names[s_g_inds]

    with measure_time('Computing scores...', verbose=verbose):
      mAP, cmc_scores = compute_score(q_g_dist,query_ids=ids[s_q_inds],
        gallery_ids=ids[s_g_inds],
        query_tags=tags[s_q_inds],
        gallery_tags=tags[s_g_inds])

    print('{:<30}'.format('Small Single Query:'), end='')
    print_scores(mAP, cmc_scores)

    ################
    # Single Query #
    ################

    with measure_time('Computing distance...', verbose=verbose):
      # query-gallery distance
      q_g_dist = compute_dist(feat[m_q_inds], feat[m_g_inds], type='euclidean')
      m_indices = np.argsort(q_g_dist, axis=1)
      m_q_names = im_names[m_q_inds]
      m_g_names = im_names[m_g_inds]

    with measure_time('Computing scores...', verbose=verbose):
      mAP, cmc_scores = compute_score(q_g_dist,query_ids=ids[m_q_inds],
        gallery_ids=ids[m_g_inds],
        query_tags=tags[m_q_inds],
        gallery_tags=tags[m_g_inds])

    print('{:<30}'.format('Medium Single Query:'), end='')
    print_scores(mAP, cmc_scores)

    ################
    # Single Query #
    ################

    with measure_time('Computing distance...', verbose=verbose):
      # query-gallery distance
      q_g_dist = compute_dist(feat[l_q_inds], feat[l_g_inds], type='euclidean')
      l_indices = np.argsort(q_g_dist, axis=1)
      l_q_names = im_names[l_q_inds]
      l_g_names = im_names[l_g_inds]

    with measure_time('Computing scores...', verbose=verbose):
      mAP, cmc_scores = compute_score(q_g_dist,query_ids=ids[l_q_inds],
        gallery_ids=ids[l_g_inds],
        query_tags=tags[l_q_inds],
        gallery_tags=tags[l_g_inds])

    print('{:<30}'.format('Large Single Query:'), end='')
    print_scores(mAP, cmc_scores)

    if to_re_rank:

      ##########################
      # Re-ranked Single Query #
      ##########################

      with measure_time('Re-ranking distance...', verbose=verbose):
        # query-query distance
        q_q_dist = compute_dist(feat[l_q_inds], feat[l_q_inds], type='euclidean')
        # gallery-gallery distance
        g_g_dist = compute_dist(feat[l_g_inds], feat[l_g_inds], type='euclidean')
        # re-ranked query-gallery distance
        re_r_q_g_dist = re_ranking(q_g_dist, q_q_dist, g_g_dist)

      with measure_time('Computing scores for re-ranked distance...',
                        verbose=verbose):
        mAP, cmc_scores = compute_score(re_r_q_g_dist)

      print('{:<30}'.format('Re-ranked Single Query:'), end='')
      print_scores(mAP, cmc_scores)

    return s_indices,s_q_names,s_g_names,m_indices,m_q_names,m_g_names,l_indices,l_q_names,l_g_names,mAP, cmc_scores
