from .Dataset import Dataset
from ..utils.dataset_utils import parse_im_name

import os.path as osp
from PIL import Image
import numpy as np
from collections import defaultdict
import math

import os


class TrainSet(Dataset):
  """Training set for triplet loss.
  Args:
    ids2labels: a dict mapping ids to labels
  """

  def __init__(
      self,
      im_dir=None,
      im_names=None,
      ids2labels=None,
      ids_per_batch=None,
      ims_per_id=None,
      **kwargs):

    # The im dir of all images
    self.im_dir = im_dir
    self.im_names = im_names
    self.ids2labels = ids2labels
    self.ids_per_batch = ids_per_batch
    self.ims_per_id = ims_per_id

    im_ids = [parse_im_name(name, 'id') for name in im_names]
    self.ids_to_im_inds = defaultdict(list)
    for ind, id in enumerate(im_ids):
      self.ids_to_im_inds[id].append(ind)
    self.ids = self.ids_to_im_inds.keys()

    window_f = os.path.abspath('/home/bing/Dataset/VehicleID/window-VehicleID.txt')
    lines = {}
    with open(window_f, 'r') as f:
      lines = {l.strip().split(' ')[0]: l.strip().split(' ')[1:6] for l in f if
               not lines.has_key(l.strip().split(' ')[0])}
    self.window = lines

    light_f = os.path.abspath('/home/bing/Dataset/VehicleID/light_pro.txt')
    lines = {}
    with open(light_f, 'r') as f:
      # lines = [[l.strip().split(' ')[0]]+l.strip().split(' ')[1:6] for l in f]
      for l in f:
        items = l.strip().split(' ')
        if float(items[1]) < 0.5:
          continue
        if lines.has_key(items[0]):
          lines[items[0]].append(map(float,items[2:]))
        else:
          lines[items[0]] = []
          lines[items[0]].append(map(float,items[2:]))
    self.light = lines

    super(TrainSet, self).__init__(
      dataset_size=len(self.ids),
      batch_size=ids_per_batch,
      **kwargs)

  def get_sample(self, ptr):
    """Here one sample means several images (and labels etc) of one id.
    Returns:
      ims: a list of images
    """
    inds = self.ids_to_im_inds[self.ids[ptr]]
    if len(inds) < self.ims_per_id:
      inds = np.random.choice(inds, self.ims_per_id, replace=True)
    else:
      inds = np.random.choice(inds, self.ims_per_id, replace=False)
    im_names = [self.im_names[ind] for ind in inds]

    ims = [np.asarray(Image.open(osp.join(self.im_dir, name)))
           for name in im_names]

    # window
    box1s = []
    for im, name in zip(ims, im_names):
      if self.window.has_key(name[:-4]):
        box1s.append(map(float,self.window[name[:-4]][1:]))
      else:
        box1s.append(-1)
    # lights
    box2s = []
    for im, name in zip(ims, im_names):
      if self.light.has_key(name[:-4]):
        box2s.append(self.light[name[:-4]])
      else:
        box2s.append(-1)

    box1, box2, box3, box4 = self.box_generate(ims,box1s, box2s)
    ims,mirrored = zip(*[self.pre_process_im(im) for im in ims])

    labels = [self.ids2labels[self.ids[ptr]] for _ in range(self.ims_per_id)]
    return ims, im_names, labels,box1,box2,box3,box4, mirrored
  def box_generate(self,ims,box1s,box2s):
    """
    :param box1s: [N,4] or [N, None]
    :param box2s: [N,k,5] or [N,None]
    :return:
    box1:[N,4]for window
    box2:[N,4]for left light
    box3:[N,4]for right light
    box4:[N,4]for brand
    """
    box1 = np.zeros((self.ims_per_id,4))
    box2 = np.zeros((self.ims_per_id, 4))
    box3 = np.zeros((self.ims_per_id, 4))
    box4 = np.zeros((self.ims_per_id, 4))
    for i in range(len(box1s)):
      if box1s[i] ==-1:
        box1[i] = np.array([0,3,15,7])
      else:
        # x [0,256],x/16 [0,16]
        x1,y1,x2,y2 = box1s[i]
        x1 = float(x1) / ims[i].shape[1] * 256
        x2 = float(x2) / ims[i].shape[1] * 256
        y1 = float(y1) / ims[i].shape[0] * 256
        y2 = float(y2) / ims[i].shape[0] * 256

        xtl = int(x1 / 16)
        xbr = int(math.ceil(x2 / 16))
        ytl = int(y1 / 16)
        ybr = int(math.ceil(y2 / 16))
        box1[i] = np.array([xtl,ytl,xbr,ybr])
    for i in range(len(box2s)):
      if box2s[i] == -1:
        box2[i] = np.array([0,11,5,16])
        box3[i] = np.array([11,11,16,16])
        box4[i] = np.array([5, 11, 11, 16])
      elif len(box2s[i]) == 1:
        x1,y1,x2,y2 = box2s[i][0]
        x1 = float(x1) / ims[i].shape[1] * 256
        x2 = float(x2) / ims[i].shape[1] * 256
        y1 = float(y1) / ims[i].shape[0] * 256
        y2 = float(y2) / ims[i].shape[0] * 256

        ytl = int(x1 / 16)
        ybr = int(math.ceil(x2 / 16))
        xtl = int(y1 / 16)
        xbr = int(math.ceil(y2 / 16))
        box2[i] = np.array([xtl, ytl, xbr, ybr])
        box3[i] = np.array([xtl, ytl, xbr, ybr])
        box4[i] = np.array([5,11,11,16])
      else :
        x1,y1,x2,y2 = box2s[i][0]
        x1 = float(x1) / ims[i].shape[1] * 256
        x2 = float(x2) / ims[i].shape[1] * 256
        y1 = float(y1) / ims[i].shape[0] * 256
        y2 = float(y2) / ims[i].shape[0] * 256

        xtl1 = int(x1 / 16)
        xbr1 = int(math.ceil(x2 / 16))
        ytl1 = int(y1 / 16)
        ybr1 = int(math.ceil(y2 / 16))
        assert xbr1<=16 and ybr1<=16, 'xbr1<16'
        box2[i] = np.array([xtl1, ytl1, xbr1, ybr1])
        x1,y1,x2,y2 = box2s[i][1]
        x1 = float(x1) / ims[i].shape[1] * 256
        x2 = float(x2) / ims[i].shape[1] * 256
        y1 = float(y1) / ims[i].shape[0] * 256
        y2 = float(y2) / ims[i].shape[0] * 256
        xtl0 = int(x1 / 16)
        xbr0 = int(math.ceil(x2 / 16))
        ytl0 = int(y1 / 16)
        ybr0 = int(math.ceil(y2 / 16))
        assert xbr0 <= 16 and  ybr0 <= 16, 'xbr0<16'
        box3[i] = np.array([xtl0, ytl0, xbr0, ybr0])

        xtl = xbr0 if xbr0 < xbr1 else xbr1
        xbr = xtl0 if xtl0 > xtl1 else xtl1
        ytl = ytl1 if ytl1 < ytl0 else ytl0
        ybr = ybr1 if ybr1 > ybr0 else ybr0

        if not (xtl<xbr and ytl < ybr):
          box4[i] = np.array([5,11,11,16])
        else:
          box4[i] = np.array([xtl,ytl,xbr,ybr])
    return box1,box2,box3,box4

  def next_batch(self):
    """Next batch of images and labels.
    Returns:
      ims: numpy array with shape [N, H, W, C] or [N, C, H, W], N >= 1
      img_names: a numpy array of image names, len(img_names) >= 1
      labels: a numpy array of image labels, len(labels) >= 1
      mirrored: a numpy array of booleans, whether the images are mirrored
      self.epoch_done: whether the epoch is over
    """
    # Start enqueuing and other preparation at the beginning of an epoch.
    if self.epoch_done and self.shuffle:
      np.random.shuffle(self.ids)
    samples, self.epoch_done = self.prefetcher.next_batch()
    im_list, im_names, labels, box1,box2,box3,box4, mirrored = zip(*samples)
    # t = time.time()
    # Transform the list into a numpy array with shape [N, ...]
    ims = np.stack(np.concatenate(im_list))
    # print '---stacking time {:.4f}s'.format(time.time() - t)
    im_names = np.concatenate(im_names)
    labels = np.concatenate(labels)
    mirrored = np.concatenate(mirrored)
    box1 = np.concatenate(box1)
    box2 = np.concatenate(box2)
    box3 = np.concatenate(box3)
    box4 = np.concatenate(box4)
    return ims, im_names, labels,box1.astype(int),box2.astype(int),box3.astype(int),box4.astype(int), mirrored, self.epoch_done
