import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from .resnet import resnet50


class Model(nn.Module):
  def __init__(self, last_conv_stride=2,
               num_classes=13164,
               num_stripes=4,
               local_conv_out_channels=256):
    super(Model, self).__init__()
    self.base = resnet50(pretrained=True, 
        last_conv_stride=last_conv_stride)

    self.num_stripes = num_stripes
    self.local_conv_list = nn.ModuleList()
    for _ in range(num_stripes):
      self.local_conv_list.append(nn.Sequential(
        nn.Conv2d(2048, local_conv_out_channels, 1),
        nn.BatchNorm2d(local_conv_out_channels),
        nn.ReLU(inplace=True)
      ))

    if num_classes > 0:
      self.fc_list = nn.ModuleList()
      for _ in range(num_stripes):
        fc = nn.Linear(local_conv_out_channels, num_classes)
        init.normal(fc.weight, std=0.001)
        init.constant(fc.bias, 0)
        self.fc_list.append(fc)
  def forward(self, x,box1,box2,box3,box4):
    # shape [N, C, H, W]
    feat = self.base(x)

    local_feat_list = []
    logits_list = []

    N, C, H, W = feat.shape
    feat_window = torch.autograd.Variable(torch.zeros((N,C,1,1)).cuda())
    feat_light = torch.autograd.Variable(torch.zeros((N,C,1,1)).cuda())
    feat_brand = torch.autograd.Variable(torch.zeros((N,C,1,1)).cuda())

    for i in range(N):
      light0 = feat[i,:,box2[i][1]:box2[i][3],box2[i][0]:box2[i][2]]
      light1 = feat[i, :,box3[i][1]:box3[i][3],box3[i][0]:box3[i][2]]
      #light = (light0+light1)/2
      if not (len(light0.size()[1:]) ==2 and len(light1.size()[1:]) ==2):
        print 'light size'
        light0 = feat[i, :, 11:16, 0:5]
        light1 = feat[i, :, 11:16,11:16]
      #assert len(light0.size()[1:]) ==2 and len(light1.size()[1:]) ==2,'light size'
      light0 = F.avg_pool2d(light0, light0.size()[1:])
      light1 = F.avg_pool2d(light1, light1.size()[1:])
      feat_light[i] = (light0+light1)/2

      brand = feat[i,:,box4[i][1]:box4[i][3],box4[i][0]:box4[i][2]]
      feat_brand[i] = F.avg_pool2d(brand, brand.size()[1:])

      #no window -> (3,7)
      try:
        window = feat[i,:,box1[i][1]:box1[i][3],box1[i][0]:box1[i][2]]
      except:
        print 'window except'
      feat_window[i] = F.avg_pool2d(window,window.size()[1:])


    global_ = F.avg_pool2d(feat,feat.size()[2:])

    for i,local_feat in enumerate([feat_light,feat_window,feat_brand,global_]):
      #shape [N, c, 1, 1]
      try:
      	local_feat = self.local_conv_list[i](local_feat)
      	# shape [N, c]
      	local_feat = local_feat.view(local_feat.size(0), -1)
      	local_feat_list.append(local_feat)
      	if hasattr(self, 'fc_list'):
        	logits_list.append(self.fc_list[i](local_feat))
      except:
        print 'conv fc'

    if hasattr(self, 'fc_list'):
      return local_feat_list, logits_list

    return local_feat_list
